package edu.stanford.rsl.tutorial.book.bookfitting.old;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import edu.stanford.rsl.conrad.data.numeric.Grid3D;

import edu.stanford.rsl.conrad.data.numeric.opencl.OpenCLGrid3D;
import edu.stanford.rsl.conrad.filtering.MedianFilteringTool;
import edu.stanford.rsl.conrad.utils.ImageUtil;
import ij.ImageJ;
import ij.ImagePlus;

import ij.io.Opener;
import ij.plugin.filter.RankFilters;
import ij.process.FloatProcessor;

/**
 * @author DanielSt
 * @date: 05/14/2018
 * 
 * Code for extraction of book pages
 * Part 1: preprocessing and iterative process w/o 2-D mapping
 */

public class PageFlattenerNNApproximation {

	//Set path of volume and Vesselness filtered volume - it should be the xz-direction!!!!!
	private String vesselnessXY ="path_to_vol";
	private String originalVolume = "path_to_vol";
	
	private Opener opener;
	private ImagePlus imp;
	private OpenCLGrid3D resultVolume;
	private OpenCLGrid3D processingVolumeYZ;
	private OpenCLGrid3D orgVolume;
	
	private MedianFilteringTool median = new MedianFilteringTool();

	//maximum page thickness
	private int maxPageThicknessInPixel = 6;
	//number of pages
	private int numberOfPages = 8;
	private long bufferSize;
	private long size = 0;
	
	//threshold for vesselness binarization
	private final float thresh = 3700f;
	//Median kernel size
	private int medianKernel = 5;

	//constructor w/o parameter
	public PageFlattenerNNApproximation(){
		//load imageJ, lead volume into array
		new ImageJ();
		
		opener = new Opener();
		imp = opener.openImage(vesselnessXY);
		
		processingVolumeYZ = new OpenCLGrid3D(ImageUtil.wrapImagePlus(imp));
		processingVolumeYZ.setSpacing(1,1,1);
		processingVolumeYZ.setOrigin(0,0,0);
		
		imp = opener.openImage(originalVolume);
		orgVolume = new OpenCLGrid3D(ImageUtil.wrapImagePlus(imp));
		orgVolume.setSpacing(1,1,1);
		orgVolume.setOrigin(0,0,0);
		
		orgVolume.show();

		bufferSize = Long.MAX_VALUE;
		//configure median filter
		median.configure(medianKernel,medianKernel);
		
		//run program
		start();
		//show result
		processingVolumeYZ.clone().show();
	}
	
	//contructor with parameter - given grid
	public PageFlattenerNNApproximation(OpenCLGrid3D grid){
		new ImageJ();

		processingVolumeYZ = grid;
		processingVolumeYZ.setSpacing(1,1,1);
		processingVolumeYZ.setOrigin(0,0,0);

		bufferSize = Long.MAX_VALUE;
		median.configure(medianKernel,medianKernel);
		start();
		processingVolumeYZ.clone().show();
	}
	
	// get result vol
	public OpenCLGrid3D getResult(){
		return processingVolumeYZ;
	}
	
	// start process
	public void start(){
		
		//time measurement
		long total = 0;
		long init = System.currentTimeMillis();
		// separate pages process
		separateTwoPages();
		
		//time
		init =System.currentTimeMillis()-init;
		System.out.println("init: "+init);
		//iterate (stop limit 50)
		for(int iter = 0; iter < 50; iter++){
			// wrongly detected rows
			size = 0;

			long time = System.currentTimeMillis();
			// approximate pages
			approximatePages();
			time = System.currentTimeMillis()-time;
			
			//rotate volume for filtering in xy direction
			resultVolume = new OpenCLGrid3D(new Grid3D(processingVolumeYZ.getSize()[2],processingVolumeYZ.getSize()[1],processingVolumeYZ.getSize()[0]));
			resultVolume.getGridOperator().rotateAroundYAxis(resultVolume, processingVolumeYZ);
			processingVolumeYZ.release();
			
			//median filtering
			long time1 = System.currentTimeMillis();
			for(int i = 0; i< resultVolume.getSize()[2];i++)
				resultVolume.setSubGrid(i,median.applyToolToImage(resultVolume.getSubGrid(i)));
			time1 = System.currentTimeMillis()-time1;
			
			//rerotate 
			processingVolumeYZ = new OpenCLGrid3D(new Grid3D(resultVolume.getSize()[2],resultVolume.getSize()[1],resultVolume.getSize()[0]));
			orgVolume.getGridOperator().rotateAroundYAxis(processingVolumeYZ, resultVolume);
			resultVolume.release();	
			total += time +time1;
			
			//wrongly detected page-rows
			if(size > bufferSize){
				System.out.println("Wrongly detected: "+size);
				break;
			}
			
			System.out.println("Wrongly detected rows: "+size);

			bufferSize = size;
			
			//threshold volume
			for(int z = 0 ; z < orgVolume.getSize()[2]; z++)
				for(int y = 0; y < orgVolume.getSize()[1]; y++)
					for(int x = 0; x < orgVolume.getSize()[0]; x++){
						if(orgVolume.getAtIndex(x, y, z) < thresh)
							processingVolumeYZ.setAtIndex(x, y, z, 0);
					}
		}

		System.out.println((total+init)/1000);
	}
	
	/**
	 * Separate two pages by the proposed algorithm
	 */
	public void separateTwoPages(){
		//outlier filter to delete small pixel errors
		for(int z = 0; z < processingVolumeYZ.getSize()[2]; z++){
			RankFilters rf = new RankFilters();
			FloatProcessor gridProc = ImageUtil.wrapGrid2D(processingVolumeYZ.getSubGrid(z));
			rf.rank(gridProc, 1f, RankFilters.OUTLIERS, -1 , 0.5f);
			processingVolumeYZ.setSubGrid(z,ImageUtil.wrapFloatProcessor(gridProc));
		}
		//go through complete volume (z), through every row (y)
		for(int z = 0 ; z < processingVolumeYZ.getSize()[2]; z++)
			for(int y = 1; y < processingVolumeYZ.getSize()[1]; y++){
				//page count and save it hash map
				int countPages = 0;
				HashMap<Integer,Integer> arr = new HashMap<Integer,Integer>();
				// go through row
				for(int x = 0; x < processingVolumeYZ.getSize()[0]; x++){
					//get thickness of pages and write in array, count pages
					int thick = 0;
					while(processingVolumeYZ.getSize()[0] > x && processingVolumeYZ.getAtIndex(x++, y, z) > 0.0f){
						thick++;
						if( x < processingVolumeYZ.getSize()[0] && processingVolumeYZ.getAtIndex(x, y, z) == 0.0f){
							countPages++;
							arr.put(x-1,thick);
							break;
						}
					}
				}
				
				// if less pages detected than given, split thickest pages if they are thicker than the maximum
				// if the are even thicker, split pages twice (multiple overlaps)
				if(countPages < numberOfPages){
					//go over all rows
					for(int x = 0; x < processingVolumeYZ.getSize()[0]; x++){
						//if the pixel value is greater than zero, calculate the diameter of the page by counting valid pixels
						if(processingVolumeYZ.getAtIndex(x, y, z) >0.0f){
							int length = 0;
	
							while(x < processingVolumeYZ.getSize()[0] && processingVolumeYZ.getAtIndex(x++, y, z) > 0.0f)
								length++;

							//if length > max allowed page thickness and less then 1.5 of the max, split once
							if(length > maxPageThicknessInPixel && length <= 1.5*maxPageThicknessInPixel){
								if(y == 0) {
									processingVolumeYZ.setAtIndex(x-1-length/2, y, z, 0);
									countPages++;
								}
								else if(	processingVolumeYZ.getAtIndex(x-1-length/2, y-1, z) == 0 
											|| processingVolumeYZ.getAtIndex(x-length/2, y-1, z) == 0
											|| processingVolumeYZ.getAtIndex(x-2-length/2, y-1, z) == 0){
									processingVolumeYZ.setAtIndex(x-1-length/2, y, z, 0);
									countPages++;
								}
							//if diameter is greater than 2.5, split twice
							} else if (length > 2.5 * maxPageThicknessInPixel)
								if (y == 0){
									processingVolumeYZ.setAtIndex(x-1-length/3, y, z, 0);
									processingVolumeYZ.setAtIndex(x-1-2*length/3, y, z, 0);
									countPages++;
								} else if(processingVolumeYZ.getAtIndex(x-3-length/2, y-1, z) == 0
									|| processingVolumeYZ.getAtIndex(x-2-length/2, y-1, z) == 0 
									|| processingVolumeYZ.getAtIndex(x-1-length/2, y-1, z) == 0 
									|| processingVolumeYZ.getAtIndex(x-length/2, y-1, z) == 0
									|| processingVolumeYZ.getAtIndex(x+1-length/2, y-1, z) == 0 ){
								processingVolumeYZ.setAtIndex(x-1-length/3, y, z, 0);
								processingVolumeYZ.setAtIndex(x-1-2*length/3, y, z, 0);
								countPages++;
								}
							}
						}
				} 
				
				//if more pages detected than given, remove it
				while(countPages > numberOfPages){
					
					int min = 20000;
					int idx = 0;
					
					Iterator iter = arr.entrySet().iterator();
					while (iter.hasNext()) {
						HashMap.Entry entry = (Map.Entry) iter.next();
						int key = (int)entry.getKey();
						int val = (int)entry.getValue();
					
						if(min > val){
							min = val;
							idx = key;
							}
					}
					for(int x = 0; x < min; x++){
						processingVolumeYZ.setAtIndex(idx-x, y, z, 0);
					}
					arr.remove(idx);
					countPages--;
				}
			}
	}
	
	// approximate pages
	public void approximatePages(){
		//saving the corrupted rows
		ArrayList<Integer> corruptedRows[] = new ArrayList[processingVolumeYZ.getSize()[2]];
		for(int z = 0; z < processingVolumeYZ.getSize()[2]; z++)
			corruptedRows[z] = getCorruptedRows(z);
		
		boolean switcher = false;
		int z = processingVolumeYZ.getSize()[2]/2-1;

		// search the nearest neighbor of a row in the volume, that actually has the defined number of pages and substitute
		//go through all slices
		while(z >= 0 && z<processingVolumeYZ.getSize()[2]){
			//walk over all rows and search the nearest neighbor - toggle switches from forward to backward
			for(int i = 0; i < corruptedRows[z].size(); i++){
				int toFit = corruptedRows[z].get(i);
				size++;
				int toggler = 1;
				int sliceToCopyFrom = -1;
				while(toggler < processingVolumeYZ.getSize()[2]){

					if(z+toggler < processingVolumeYZ.getSize()[2] && !corruptedRows[z+toggler].contains(toFit))
							sliceToCopyFrom=z+toggler;
					
					if(z-toggler >= 0 && !corruptedRows[z-toggler].contains(toFit))
							sliceToCopyFrom=z-toggler;
					// store the missing line into the volume
					if(sliceToCopyFrom > 0){
						fillMissingLines(toFit, sliceToCopyFrom, z);
						break;
					}
					
					toggler++;
				}
			}
			//switch forward to backwards
			if(!switcher)
				z--;
			else 
				z++;
			if(z <0){
				z = processingVolumeYZ.getSize()[2]/2;
				switcher = true;
			}
		}

	}
	
	// fit in a row at a given position in the volume
	public void fillMissingLines(int toFit, int sliceToCopyFrom, int z){
		for(int x = 0; x < processingVolumeYZ.getSize()[0]; x++)
			processingVolumeYZ.setAtIndex(x, toFit, z, processingVolumeYZ.getAtIndex(x,toFit,sliceToCopyFrom));
	}
	
	//get all corrupted rows
	public ArrayList<Integer> getCorruptedRows(int z){
		ArrayList<Integer> corruptedRows = new ArrayList<Integer>();
		for(int y = 0; y < processingVolumeYZ.getSize()[1]; y++){
			float ctr = 0;
			for(int x = 0; x < processingVolumeYZ.getSize()[0]; x++){
				if(processingVolumeYZ.getAtIndex(x, y, z)==255){
					ctr++;
					while(x < processingVolumeYZ.getSize()[0] && processingVolumeYZ.getAtIndex(x++, y, z)==255);
				}
			}
			//if break
			if(ctr != numberOfPages){			
				for(int x = 0; x < processingVolumeYZ.getSize()[0]; x++){
					processingVolumeYZ.setAtIndex(x, y, z, 0);
				}
				corruptedRows.add(y);
			}
		}

		return corruptedRows;
	}
	
	//sanity check to count the initial number of detcted pages - automatic page number detection by averaging
	public void sanityCheck(int z){
		float counter = 0 ;
		for(int y = 0; y < processingVolumeYZ.getSize()[1]; y++){
			for(int x = 0; x < processingVolumeYZ.getSize()[0]; x++){
				if(processingVolumeYZ.getAtIndex(x, y, z)==255){
					counter++;
					while(processingVolumeYZ.getAtIndex(x++, y, z)==255);
				}
			}
		}
		numberOfPages = Math.round(counter/processingVolumeYZ.getSize()[1]);
		System.out.println("Pages detected: "+numberOfPages);

	}

	public static void main(String[] args){
		PageFlattenerNNApproximation  runIt = new PageFlattenerNNApproximation ();
	}

}